/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#ifndef _XMA_LOGGER_H_
#define _XMA_LOGGER_H_

#include <stdint.h>
#include <stddef.h>

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Alias XmaLogLevelType to standard SysLog levels
 */

typedef enum XmaLogLevelType {
  XMA_EMERGENCY_LOG = 0,
  XMA_ALERT_LOG     = 1,
  XMA_CRITICAL_LOG  = 2,
  XMA_ERROR_LOG     = 3,
  XMA_WARNING_LOG   = 4,
  XMA_NOTICE_LOG    = 5,
  XMA_INFO_LOG      = 6,
  XMA_DEBUG_LOG     = 7
} XmaLogLevelType;

typedef enum XmaLogType {
  XMA_LOG_TYPE_NONE,
  XMA_LOG_TYPE_CONSOLE,
  XMA_LOG_TYPE_SYSLOG,
  XMA_LOG_TYPE_FILE,
  XMA_LOG_TYPE_CALLBACK,
} XmaLogType;

typedef void (*XmaLogCallbackFunction)(void* opaque, XmaLogLevelType level, const char* name, const char* msg);

typedef void* XmaLogHandle;

/**
 * xma_log_init() - This function initializes the logs.
 *
 * @log_level:     Minimum logging level for a message to be added to the log
 * @log_type:      Location to write the log to
 * @handle:        Pointer to recieve handle
 * ...:            If the log_type is 'XMA_LOG_TYPE_FILE', this will be a
 *                 character string that specifies the file to log to. If the
 *                 log_type is 'XMA_LOG_TYPE_CALLBACK', this will be a function
 *                 pointer of type XmaLogCallbackFunction followed by a void*
 *                 parameter that will be passed to the callback as opaque
 *
 * RETURN: XMA_SUCCESS on success
 *         XMA_ERROR_BAD_ALLOC on out of memory
 *         XMA_ERROR_INVALID on invalid input
*/
int32_t xma_log_init(XmaLogLevelType log_level, XmaLogType log_type, XmaLogHandle* handle, ...);

/**
 * xma_release_logging() - This function releases memory associated with the
 * logs
 *
 * @handle: Handle to log to release
 *
*/
void xma_log_release(XmaLogHandle handle);

/**
 * xma_logmsg() - This function logs a message to the log handle. The log
 * message includes the current time, logging level, a unique name, and a
 * message.
 *
 * @handle: Handle to XMA logging
 * @level:  Logging level associated with the message
 * @name:   Pointer to a C string indicating the name of the entity that is
 *          generating the log message
 * @msg:    Pointer to a compatible C format string
 * @...:    Variable arguments used in conjunction with the C format string
 *          contained in the msg parameter
 *
*/
void xma_logmsg(XmaLogHandle handle, XmaLogLevelType level, const char* name, const char* msg, ...);

#ifdef __cplusplus
}
#endif

#endif
