/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#include <memory>
#include <cstring>
#include "InternalDecoderSession.h"
#include "InternalDataBuffer.h"

namespace xma {

#define XMA_DECODER_SESSION "DecoderSession"

/* The delay will be relative to the frame rate, so a framerate of 60 fps and
 * retry delay of 10 will lead to a delay of 1/(60*10) seconds */
#define XMA_DECODER_RETRY_DELAY 10
/* The retry count is relative to the retry delay (which is relative to the
 * framerate), so if the framerate is 60 fps and the retry count is
 * (XMA_DECODER_RETRY_DELAY*10), the maximum delay will be 10*1/60 or 1/6
 * seconds */
#define XMA_DECODER_MAX_RETRY_COUNT (XMA_DECODER_RETRY_DELAY * 10)

#define MAX_DECODER_PARAMETERS 1
const char* decoder_parameter_names[MAX_DECODER_PARAMETERS][2] = {
    {"enable_mc", "enable_mc"},
};

///////////////////////////////////////////////////////////////////////////////
InternalDecoderSession::InternalDecoderSession(IntSessionFactoryPtr sf, XmaDecoderProperties* props) : SessionBase(sf) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession()");

  m_mem_ctx = vpi_mem_handle_t(m_sf->CreateMemHandle(), [](VpiMemHandle mem_ctx) { vpi_memory_release(mem_ctx); });
  if (!m_mem_ctx) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Failed to create VPI memory!");
  }

  VpiDecInitParams params;
  memset(&params, 0, sizeof(VpiDecInitParams));
  VpiHwId hw_id;
  switch (props->hwdecoder_type) {
  case XMA_H264_DECODER_TYPE:
    hw_id           = VPE_HW_H264DEC;
    params.dec_type = VPE_DEC_H264;
    break;
  case XMA_HEVC_DECODER_TYPE:
    hw_id           = VPE_HW_HEVCDEC;
    params.dec_type = VPE_DEC_HEVC;
    break;
  case XMA_JPEG_DECODER_TYPE:
    hw_id           = VPE_HW_JPEGDEC;
    params.dec_type = VPE_DEC_JPEG;
    break;
  case XMA_VP9_DECODER_TYPE:
    hw_id           = VPE_HW_VP9DEC;
    params.dec_type = VPE_DEC_VP9;
    break;
  case XMA_AV1_DECODER_TYPE:
    hw_id           = VPE_HW_AV1DEC;
    params.dec_type = VPE_DEC_AV1;
    break;
  default:
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Invalid decoder type!");
  }
  params.dev_ctx    = m_sf->GetDevHandle();
  params.mem_ctx    = m_mem_ctx.get();
  params.src_width  = props->width;
  params.src_height = props->height;
  params.resize_str = nullptr;
  params.src_depth  = props->bits_per_pixel;

  VpiHwCfg* hw_cfg = vpi_device_get_hw_config(m_sf->GetDevHandle(), hw_id);

  if (hw_cfg->sw_formats[0] == VPI_FORMAT_NONE) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Hardware error format!");
  }

  if (hw_cfg->hw_formats[0] != VPI_FORMAT_VPE) {
    THROW_INVALID_ARGUMENT_EXCEPTION(
        m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Hardware does not offer any usable surface formats!");
  }
  vpi_device_free_hw_config(hw_cfg);

  params.transcode = 0;

  VpiBufferPool* new_pool = nullptr;
  VpiRet         ret      = vpi_buffer_pool_create(m_sf->GetMemHandle(), XMA_DEFAULT_PKT_POOL_SIZE, XMA_DEFAULT_PKT_SIZE, RC_SIDE_EN, false, 0, &new_pool);
  if (ret < VPI_SUCCESS) {
    THROW_RUNTIME_EXCEPTION(
        m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Failed to create buffer pool: %s!", vpi_error_str(ret));
  }
  m_pool          = pool_handle_t(new_pool, [this](void* pool) {
    if (pool) {
      DataBuffer* buf = static_cast<DataBuffer*>(vpi_decoder_flush_pkt(pool));
      while (buf) {
        lock_t guard(m_mutex);
        m_buffer_map.erase(buf);
        guard.unlock();
        buf = static_cast<DataBuffer*>(vpi_decoder_flush_pkt(pool));
      }
      vpi_pool_unref(pool);
    }
  });
  params.pkt_pool = m_pool.get();

  XmaParameter* param;
  if ((param = GetParameter(props->params, props->param_cnt, XMA_DEC_PARAM_LOW_LATENCY)) != nullptr) {
    if (*(int32_t*) param->value < 0 || *(int32_t*) param->value > 1) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Invalid value for low_latency");
    } else {
      m_low_latency = static_cast<bool>(*(int32_t*) param->value);
    }
  } else {
    m_low_latency = XMA_DEC_LOW_LATENCY_DEFAULT;
  }
  params.low_latency = m_low_latency;

  if ((param = GetParameter(props->params, props->param_cnt, XMA_DEC_PARAM_LATENCY_LOGGING)) != nullptr) {
    if (*(int32_t*) param->value < 0 || *(int32_t*) param->value > 1) {
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Invalid value for latency_logging");
    } else {
      m_latency_logging = *(int32_t*) param->value;
    }
  } else {
    m_latency_logging = XMA_DEC_LATENCY_LOGGING_DEFAULT;
  }

  const char* fmt = nullptr;
  param           = GetParameter(props->params, props->param_cnt, XMA_DEC_OUTPUT_FORMAT);
  if (param) {
    switch (*static_cast<int32_t*>(param->value)) {
    case XMA_NV12_FMT_TYPE:
      if (props->bits_per_pixel != 8) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION,
            "InternalDecoderSession::InternalDecoderSession() Input bit depth (%d) and output bit depth (8) do not match!", props->bits_per_pixel);
      }
      if (props->flags & XMA_FRAME_PROPERTY_FLAG_TILE_4x4) {
        if (props->flags & XMA_FRAME_PROPERTY_FLAG_COMPRESS) {
          fmt = "nv12-ma";
        } else {
          fmt = "nv12-tile";
        }
      } else {
        if (props->flags & XMA_FRAME_PROPERTY_FLAG_COMPRESS) {
          THROW_INVALID_ARGUMENT_EXCEPTION(
              m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() NV12 pixel format does not support compression without tile!");
        } else {
          fmt = "nv12";
        }
      }
      break;
    case XMA_P010LE_FMT_TYPE:
      if (props->bits_per_pixel != 10) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION,
            "InternalDecoderSession::InternalDecoderSession() Input bit depth (%d) and output bit depth (10) do not match!", props->bits_per_pixel);
      }
      if (props->flags & XMA_FRAME_PROPERTY_FLAG_TILE_4x4) {
        if (props->flags & XMA_FRAME_PROPERTY_FLAG_COMPRESS) {
          fmt = "p010le-ma";
        } else {
          fmt = "p010le-tile";
        }
      } else {
        if (props->flags & XMA_FRAME_PROPERTY_FLAG_COMPRESS) {
          THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION,
              "InternalDecoderSession::InternalDecoderSession() P010LE pixel format does not support compression without tile!");
        } else {
          fmt = "p010le";
        }
      }
      break;
    case XMA_PACKED10_FMT_TYPE:
      if (props->bits_per_pixel != 10) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION,
            "InternalDecoderSession::InternalDecoderSession() Input bit depth (%d) and output bit depth (10) do not match!", props->bits_per_pixel);
      }
      if (props->flags & (XMA_FRAME_PROPERTY_FLAG_TILE_4x4 | XMA_FRAME_PROPERTY_FLAG_COMPRESS)) {
        THROW_INVALID_ARGUMENT_EXCEPTION(
            m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Packed10 pixel format does not support compression or tile!");
      }
      fmt = "packed10";
      break;
    case XMA_RGB24_P_FMT_TYPE:
      if (props->flags & (XMA_FRAME_PROPERTY_FLAG_TILE_4x4 | XMA_FRAME_PROPERTY_FLAG_COMPRESS)) {
        THROW_INVALID_ARGUMENT_EXCEPTION(
            m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() RGBP pixel format does not support compression or tile!");
      }
      fmt = "rgbp";
      break;
    case XMA_YUV420P_FMT_TYPE:
      if (props->bits_per_pixel != 8) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION,
            "InternalDecoderSession::InternalDecoderSession() Input bit depth (%d) and output bit depth (8) do not match!", props->bits_per_pixel);
      }
      if (props->flags & (XMA_FRAME_PROPERTY_FLAG_TILE_4x4 | XMA_FRAME_PROPERTY_FLAG_COMPRESS)) {
        THROW_INVALID_ARGUMENT_EXCEPTION(
            m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() YUV420P pixel format does not support compression or tile!");
      }
      fmt = "yuv420p";
      break;
    case XMA_YUV420P10LE_FMT_TYPE:
      if (props->bits_per_pixel != 10) {
        THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION,
            "InternalDecoderSession::InternalDecoderSession() Input bit depth (%d) and output bit depth (10) do not match!", props->bits_per_pixel);
      }
      if (props->flags & (XMA_FRAME_PROPERTY_FLAG_TILE_4x4 | XMA_FRAME_PROPERTY_FLAG_COMPRESS)) {
        THROW_INVALID_ARGUMENT_EXCEPTION(
            m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() YUV420P10LE pixel format does not support compression or tile!");
      }
      fmt = "yuv420p10le";
      break;
    default:
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Unsupported output pixel format type!");
    }
  } else {
    if (props->bits_per_pixel == 8) {
      fmt = "nv12-ma";
    } else if (props->bits_per_pixel == 10) {
      fmt = "p010le-ma";
    } else {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Unsupported bits per pixel!");
    }
  }

  params.outfmt_str = const_cast<char*>(fmt);

  lock_t guard(m_mutex);

  m_handle = dec_handle_t(vpi_decoder_init(&params), [](VpiDecHandle handle) { vpi_decoder_close(handle); });
  if (!m_handle.get()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Failed to initialize decoder!");
  }
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() dec_handle : %p", m_handle.get());
  try {
    SetParams(m_handle.get(), props->params, props->param_cnt, decoder_parameter_names, MAX_DECODER_PARAMETERS, vpi_decoder_set_param, XMA_DECODER_SESSION,
        "InternalDecoderSession::InternalDecoderSession() Unable to set decoder parameter %s = \"%s\"",
        "InternalDecoderSession::InternalDecoderSession() Unknown parameter data type %d");
  } catch (...) {
    m_handle.reset();
    throw;
  }

  param = GetParameter(props->params, props->param_cnt, XMA_DEC_PARAM_WAIT);
  if (param) {
    m_wait = *static_cast<int32_t*>(param->value);
  }

  if (vpi_decoder_open(m_handle.get()) < 0) {
    m_handle.reset();
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Unable to open decoder for processing!");
  }

  m_pkt = pkt_handle_t(static_cast<VpiPkt*>(vpi_memory_get_struct(VPE_PACKET)), [](VpiPkt* pkt) { vpi_memory_free_struct(pkt); });

  VpiFrm* pic = nullptr;
  if ((vpi_frame_alloc(m_sf->GetMemHandle(), &pic) < 0) || !pic) {
    m_handle.reset();
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() Out of memory!");
  }
  InitVpiFrm(pic);
  m_pic       = pic_handle_t(pic, [sf](VpiFrm* _pic) { vpi_frame_free(sf->GetMemHandle(), _pic); });
  m_framerate = props->framerate;
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::InternalDecoderSession() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalDecoderSession::~InternalDecoderSession() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::~InternalDecoderSession()");

  if (!m_eos_pushed) {
    m_log->LogMsg(XMA_WARNING_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::~InternalDecoderSession() - End of stream not sent!");
  } else if (!m_eos_pulled) {
    m_log->LogMsg(XMA_WARNING_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::~InternalDecoderSession() - End of stream not recieved!");
  }

  m_handle.reset();
  m_pool.reset();
}

///////////////////////////////////////////////////////////////////////////////
FramePropertiesPtr InternalDecoderSession::GetProps() const {
  lock_t guard(m_mutex);

  return m_props;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalDecoderSession::SetLog(LogPtr log) {
  lock_t guard(m_mutex);
  SessionBase::SetLog(log);
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalDecoderSession::Enqueue(queue_t& data) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue()");
  DataBufferPtr buf = std::move(std::get<QueueDataBuffer>(data).data);

  if (!m_handle.get()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue() Decoder not initialized!");
  }

  lock_t guard(m_mutex);

  bool wait = false;
  if (buf) {
    m_timestamp_map[m_frame_sent] = buf->GetPts();
    m_pkt->buffer                 = buf->GetBuffer();
    m_pkt->size                   = buf->GetSize();
    m_pkt->pts                    = XMA_AV_NOPTS_VALUE;
    m_pkt->dts                    = m_frame_sent;
    m_pkt->opaque                 = buf.get();
    m_buffer_map[buf.get()]       = buf;
    buf.reset();
    if (m_latency_logging) {
      clock_gettime(CLOCK_REALTIME, &m_latency);
      m_time_taken = (m_latency.tv_sec * 1e3) + (m_latency.tv_nsec / 1e6);
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue() : %p : xma_dec_frame_sent : %lld : %lld\n", m_handle.get(), m_frame_sent + 1,
          m_time_taken);
    }
  } else {
    m_eos_pushed = true;
    wait         = true;
  }

  if (wait || (m_pkt->size > 0)) {
    int32_t ret = vpi_decoder_put_packet(m_handle.get(), m_pkt.get());
    if ((ret < 0) || ((ret == 0) && (!wait))) {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue() Failed to send data to decoder!");
    }
    m_pkt->size -= ret;
  }

  m_frame_sent++;

  if (!m_props) {
    m_first_frame = GetFrame();
    if (m_low_latency) {
      if (!m_eos_pushed) {
        while (!m_first_frame) {
          usleep(1000);
          m_first_frame = GetFrame();
        }
      }
    } else if (!m_first_frame) {
      if (wait) {
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue() - Returning try again");
        return XMA_TRY_AGAIN;
      } else {
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue() - Returning send more data");
        return XMA_SEND_MORE_DATA;
      }
    }
    m_props = m_first_frame->GetProps();
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Enqueue() - Returning success");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalDecoderSession::Dequeue(queue_t& data) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue()");
  if (!m_handle.get()) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() Decoder not initialized!");
  }

  data = QueueFrame();

  lock_t guard(m_mutex);

  FramePtr frame;
  if (m_first_frame) {
    frame = std::move(m_first_frame);
    m_first_frame.reset();
  } else if (m_wait) {
    int32_t count = 0;
    while (1) {
      frame = GetFrame();
      if (frame) {
        break;
      }
      if (m_eos_found) {
        break;
      }
      usleep(static_cast<uint32_t>(1000000L * m_framerate.denominator / (m_framerate.numerator * XMA_DECODER_RETRY_DELAY)));
      if (++count > XMA_DECODER_MAX_RETRY_COUNT) {
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() - Returning try again");
        return XMA_TRY_AGAIN;
      }
    }
  } else {
    frame = GetFrame();
  }

  guard.unlock();

  std::get<QueueFrame>(data).frame = frame;

  if (!frame) {
    if (m_eos_found) {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() - Returning EOS");
      m_eos_pulled = true;
      return XMA_EOS;
    } else {
      m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() - Returning resend and recieve");
      return XMA_RESEND_AND_RECV;
    }
  }
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() - Returning success");
  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
IntFramePtr InternalDecoderSession::GetFrame() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame()");
  DataBuffer* buf = static_cast<DataBuffer*>(vpi_decoder_release_pkt(m_handle.get()));
  while (buf) {
    m_buffer_map.erase(buf);
    buf = static_cast<DataBuffer*>(vpi_decoder_release_pkt(m_handle.get()));
  }

  int32_t ret = vpi_decoder_get_frame(m_handle.get());
  if (ret == VPI_DEC_FRAME) {
    ret = vpi_decoder_get_output_frame(m_handle.get(), m_pic.get());
    if (ret) {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() Failed to decode frame!");
    }
    int64_t dts = m_pic->dts;
    m_pic->pts  = m_timestamp_map[dts];
    m_timestamp_map.erase(dts);
    IntFramePtr frame = std::make_shared<InternalFrame>(m_sf, m_log, m_pic.release(), m_framerate);

    VpiFrm* pic = nullptr;
    if ((vpi_frame_alloc(m_sf->GetMemHandle(), &pic) < 0) || !pic) {
      THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() Out of memory!");
    }
    InitVpiFrm(pic);
    IntSessionFactoryPtr sf = m_sf;
    m_pic                   = pic_handle_t(pic, [sf](VpiFrm* _pic) { vpi_frame_free(sf->GetMemHandle(), _pic); });

    m_frame_recv++;
    if (m_latency_logging) {
      clock_gettime(CLOCK_REALTIME, &m_latency);
      m_time_taken = (m_latency.tv_sec * 1e3) + (m_latency.tv_nsec / 1e6);
      m_log->LogMsg(
          XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::Dequeue() : %p : xma_dec_frame_recv : %lld : %lld\n", m_handle.get(), m_frame_recv, m_time_taken);
    }

    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() - Returning frame");
    return frame;
  } else if (ret == VPI_DEC_EOF) {
    m_eos_found = true;
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() - Returning EOS");
    return nullptr;
  } else if (ret == -1) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() Failed to get frame!");
  } else {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_DECODER_SESSION, "InternalDecoderSession::GetFrame() - Returning NULL");
    return nullptr;
  }
}

} // namespace xma
