/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#include "InternalMioFilterSession.h"

namespace xma {

#define XMA_MIO_FILTER_SESSION "MioFilterSession"

///////////////////////////////////////////////////////////////////////////////
InternalMioFilterSession::InternalMioFilterSession(IntSessionFactoryPtr sf, XmaMioFilterProperties* props) : FilterThread<>(sf, props->params, props->param_cnt, 0, 1) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::InternalMioFilterSession()");

  m_mem_ctx = vpi_mem_handle_t(m_sf->CreateMemHandle(), [](VpiMemHandle mem_ctx) { vpi_memory_release(mem_ctx); });
  if (!m_mem_ctx) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::InternalMioFilterSession() Failed to create VPI memory!");
  }

  if (props->inputs[0].format != XMA_VPE_FMT_TYPE) {
    Close();
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::InternalMioFilterSession() Input must be on device!");
  }
  if (((props->inputs[0].width % 2) != 0) || ((props->inputs[0].height % 2) != 0)) {
    Close();
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_MIO_FILTER_SESSION,
        "InternalMioFilterSession::InternalMioFilterSession() Input size(%dx%d) must be multiple of 2!", props->inputs[0].width, props->inputs[0].height);
  }

  m_framerate = props->inputs[0].framerate;
  for (int32_t idx = 1; idx < props->num_inputs; ++idx) {
    if ((props->inputs[idx].framerate.numerator != m_framerate.numerator) || (props->inputs[idx].framerate.denominator != m_framerate.denominator)) {
      Close();
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::InternalMioFilterSession() All inputs must be same framerate!");
    }
  }

  VpiProcInitParams params;
  memset(&params, 0, sizeof(VpiProcInitParams));
  params.dev_ctx         = m_sf->GetDevHandle();
  params.mem_ctx         = m_mem_ctx.get();
  params.mem_ctx_in      = params.mem_ctx;
  params.width           = props->inputs[0].width;
  params.height          = props->inputs[0].height;
  params.input_format    = InternalFrameProperties::GetPixFmtName(props->inputs[0].format);
  params.input_sw_format = InternalFrameProperties::GetPixFmtName(props->inputs[0].sw_format);
  params.flag            = props->inputs[0].flags;
  params.nb_frames       = props->num_inputs;

  XmaParameter* param;
  switch (props->hwfilter_type) {
  case XMA_OVERLAY_FILTER_TYPE:
    params.processor_type = VPE_PROCESSOR_OVERLAY;
    param                 = GetParameter(props->params, props->param_cnt, XMA_MIOFILTER_OVERLAY_PARAM_X);
    if (param) {
      params.x_position = *static_cast<int32_t*>(param->value);
    }
    param = GetParameter(props->params, props->param_cnt, XMA_MIOFILTER_OVERLAY_PARAM_Y);
    if (param) {
      params.y_position = *static_cast<int32_t*>(param->value);
    }
    break;
  case XMA_TILE_FILTER_TYPE:
    params.processor_type = VPE_PROCESSOR_TILE;
    param                 = GetParameter(props->params, props->param_cnt, XMA_MIOFILTER_TILE_PARAM_X);
    if (param) {
      params.grid_x = *static_cast<int32_t*>(param->value);
    } else {
      params.grid_x = 2;
    }
    param = GetParameter(props->params, props->param_cnt, XMA_MIOFILTER_TILE_PARAM_Y);
    if (param) {
      params.grid_y = *static_cast<int32_t*>(param->value);
    } else {
      params.grid_y = 2;
    }
    if (params.nb_frames > params.grid_x * params.grid_y) {
      Close();
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_MIO_FILTER_SESSION,
          "InternalMioFilterSession::InternalMioFilterSession() The grid (%dx%d) is not large enough for %d inputs!", params.grid_x, params.grid_y, params.nb_frames);
    }
    break;
  case XMA_ML_FILTER_TYPE:
    params.processor_type = VPE_PROCESSOR_VIP_NBG;
    params.logical_dev    = 0;
    param                 = GetParameter(props->params, props->param_cnt, XMA_MIOFILTER_ML_CONFIG_JSON);
    if (param) {
      params.model_filename = static_cast<char*>(param->value);
    }
    break;
  case XMA_BLEND_FILTER_TYPE:
    params.processor_type = VPE_PROCESSOR_BLEND;
    if (props->num_inputs != 3) {
      Close();
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::InternalMioFilterSession() Blend filter requires 3 inputs!");
    }
    break;
  default:
    Close();
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::InternalMioFilterSession() Invalid filter type!");
  }
  m_type = props->hwfilter_type;

  int32_t output;
  for (output = 0; output < props->num_outputs; ++output) {
    if (((props->outputs[output].width % 2) != 0) || ((props->outputs[output].height % 2) != 0)) {
      Close();
      THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_MIO_FILTER_SESSION,
          "InternalMioFilterSession::InternalMioFilterSession() Output size(%dx%d) must be multiple of 2!", props->outputs[output].width, props->outputs[output].height);
    }

    m_props.push_back(std::make_shared<InternalFrameProperties>(m_sf, m_log, props->outputs[output].format, props->outputs[output].sw_format,
        props->outputs[output].width, props->outputs[output].height, props->outputs[output].flags));

    if (props->outputs[output].format != XMA_VPE_FMT_TYPE) {
      Close();
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::InternalMioFilterSession() All outputs must be on device!");
    }
    if ((props->outputs[output].framerate.numerator != props->inputs[0].framerate.numerator) ||
        (props->outputs[output].framerate.denominator != props->inputs[0].framerate.denominator)) {
      Close();
      THROW_INVALID_ARGUMENT_EXCEPTION(
          m_log, XMA_ERROR_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::InternalMioFilterSession() All outputs must be same framerate as source!");
    }
  }

  VpiProcHandle new_handle = nullptr;
  VpiRet        ret        = vpi_processor_init(&params, &new_handle);
  if (ret < VPI_SUCCESS) {
    Close();
    THROW_RUNTIME_EXCEPTION(
        m_log, XMA_ERROR_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::InternalMioFilterSession() Failed to initialize filter: %s!", vpi_error_str(ret));
  }
  m_handle = filter_handle_t(new_handle, [](void* handle) { vpi_processor_close(handle); });

  if (props->hwfilter_type == XMA_ML_FILTER_TYPE) {
    param = GetParameter(props->params, props->param_cnt, XMA_MIOFILTER_ML_PERFORMANCE);
    if (param) {
      vpi_processor_set_param(m_handle.get(), "perf", static_cast<char*>(param->value));
    }
    param = GetParameter(props->params, props->param_cnt, XMA_MIOFILTER_ML_TENSOR_DUMP);
    if (param) {
      vpi_processor_set_param(m_handle.get(), "tensor_dump", static_cast<char*>(param->value));
    }
    param = GetParameter(props->params, props->param_cnt, XMA_MIOFILTER_ML_INF_REPEAT);
    if (param) {
      vpi_processor_set_param(m_handle.get(), "inference_loop_repeat_count", static_cast<char*>(param->value));
    }
  }
  if (vpi_processor_open(m_handle.get()) < 0) {
    m_handle.reset();
    Close();
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::InternalMioFilterSession() Unable to open filter for processing");
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::InternalMioFilterSession() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalMioFilterSession::~InternalMioFilterSession() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::~InternalMioFilterSession()");

  Close();

  m_handle.reset();
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalMioFilterSession::Enqueue(queue_t& data) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::Enqueue()");

  int32_t ret = FilterThread::Push(data);

  switch (ret) {
  case XMA_TRY_AGAIN:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::Enqueue() - Returning try again");
    break;
  case XMA_SUCCESS:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::Enqueue() - Returning success");
    break;
  default:
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::Enqueue() - Returning");
    break;
  }
  return ret;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalMioFilterSession::Dequeue(queue_t& data) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::Dequeue()");

  int32_t ret = FilterThread::Pull(data);

  char error_msg[1024];
  sprintf(error_msg, "InternalMioFilterSession::Dequeue() - returning %d", ret);
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_MIO_FILTER_SESSION, error_msg);

  return ret;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalMioFilterSession::Process(queue_t& in, queue_t& out, EmptyDynamicParams*) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::Process()");
  std::vector<FramePtr> frames = std::move(std::get<QueueFrameList>(in).frame_list);
  if (frames.empty()) {
    out = QueueFrameList();
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::Process() - Returning EOS");
    return XMA_EOS;
  }

  using vpi_frm_t = std::unique_ptr<VpiFrm, std::function<void(VpiFrm*)>>;
  vpi_frm_t            first_pic;
  VpiFrm*              last_pic = nullptr;
  IntSessionFactoryPtr sf       = m_sf;
  for (uint32_t src = 0; src < frames.size(); src++) {
    VpiFrm* pic = nullptr;
    if ((vpi_frame_alloc(m_sf->GetMemHandle(), &pic) < 0) || !pic) {
      THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::Process() Out of memory!");
    }
    InitVpiFrm(pic);
    vpi_frm_t in_pic = vpi_frm_t(pic, [sf](VpiFrm* frm) {
      do {
        VpiFrm* next = static_cast<VpiFrm*>(frm->next);
        vpi_frame_unref(frm);
        if (vpi_buffer_get_ref_count(static_cast<VpiBuffer*>(frm->vpi_picinfo_buf)) == 0) {
          vpi_frame_free(sf->GetMemHandle(), frm);
        }
        frm = next;
      } while (frm);
    });

    int32_t ret;
    if ((ret = AddFrameToVpiFrm(m_sf, m_log, frames[src], in_pic.get())) != XMA_SUCCESS) {
      THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::Process() Failed to convert XmaFrame to VpiFrm!");
    }
    if (!last_pic) {
      first_pic = std::move(in_pic);
      last_pic  = first_pic.get();
    } else {
      last_pic->next = in_pic.release();
      last_pic       = static_cast<VpiFrm*>(last_pic->next);
    }
  }

  VpiFrm* pic = nullptr;
  if ((vpi_frame_alloc(m_sf->GetMemHandle(), &pic) < 0) || !pic) {
    THROW_BAD_ALLOC_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::Process() Out of memory!");
  }
  InitVpiFrm(pic);
  vpi_frm_t out_pic = vpi_frm_t(pic, [sf](VpiFrm* frm) {
    vpi_frame_unref(frm);
    if (vpi_buffer_get_ref_count(static_cast<VpiBuffer*>(frm->vpi_picinfo_buf)) == 0) {
      vpi_frame_free(sf->GetMemHandle(), frm);
    }
  });

  if (vpi_processor_process_frame(m_handle.get(), first_pic.get(), out_pic.get()) < 0) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::Process() Failed to process frame in filter!");
  }

  out                                      = QueueFrameList();
  std::get<QueueFrameList>(out).frame_list = std::vector<FramePtr>(m_props.size());

  uint32_t output;
  switch (m_type) {
  case XMA_OVERLAY_FILTER_TYPE:
  case XMA_BLEND_FILTER_TYPE:
  case XMA_TILE_FILTER_TYPE:
  case XMA_ML_FILTER_TYPE:
    first_pic.reset();
    out_pic->valid = 1; // GC620 doesn't properly set this
    for (output = 0; output < m_props.size(); ++output) {
      IntFramePtr out_frame = std::make_shared<InternalFrame>(m_sf, m_log, out_pic.release(), m_framerate, m_props[output]->GetFormat(), m_props[output]->GetSwFormat(),
          m_props[output]->GetWidth(), m_props[output]->GetHeight());
      // To Do: Remove this once the filter passes this side data
      SideDataPtr sd = frames[0]->GetSideData(XMA_FRAME_SIDE_DATA_HDR10_PARAMS);
      if (sd && (!out_frame->GetSideData(XMA_FRAME_SIDE_DATA_HDR10_PARAMS))) {
        out_frame->AddSideData(sd);
      }

      std::get<QueueFrameList>(out).frame_list[output] = out_frame;
    }
    break;
  default:
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::Process() Invalid filter type!");
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_MIO_FILTER_SESSION, "InternalMioFilterSession::Process() - Returning success");
  return XMA_SUCCESS;
}

} // namespace xma
